#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>
#include <fcall.h>
#include <9pclient.h>
#include <thread.h>
#include <cursor.h>
#include <mouse.h>
#include <keyboard.h>
#include <frame.h>
#include <plumb.h>
#include <html.h>
#include "dat.h"
#include "fns.h"

Url *
urlalloc(Runestr *src, Runestr *post, int m)
{
	Url *u;

	u = emalloc(sizeof(Url));
	copyrunestr(&u->src, src);
	if(m==HPost)
		copyrunestr(&u->post, post);
	u->method = m;
	incref(&u->ref);
	return u;
}

void
urlfree(Url *u)
{
	if(u==nil)
		return;

	if(decref(&u->ref)==0){
		closerunestr(&u->src);
		closerunestr(&u->act);
		closerunestr(&u->post);
		closerunestr(&u->ctype);
		free(u);
	}
}

Url *
urldup(Url *a)
{
	Url *b;

	b = emalloc(sizeof(Url));
	b->method = a->method;
	copyrunestr(&b->src, &a->src);
	copyrunestr(&b->act, &a->act);
	copyrunestr(&b->post, &a->post);
	copyrunestr(&b->ctype, &a->ctype);
	return b;
}

static 
Runestr
getattr(int conn, char *s)
{
	char buf[BUFSIZE];
	CFid *fid;
	int n;

	snprint(buf, sizeof(buf), "%d/%s", conn, s);
	fid = fsopen(webfs, buf, OREAD);
	if(fid == nil)
		error("can't open attr file");

	n = fsread(fid, buf, sizeof(buf)-1);
	if(n < 0)
		error("can't read");

	fsclose(fid);
	buf[n] = '\0';
	return (Runestr){runesmprint("%s", buf), n};
}

int
urlopen(Url *u)
{
	char buf[BUFSIZE];
	CFid *cfid, *fid;
	int fd, conn, n;

	cfid = fsopen(webfs, "clone", ORDWR);
	if(cfid == nil)
		error("can't open clone file");

	n = fsread(cfid, buf, sizeof(buf)-1);
	if(n <= 0)
		error("reading clone");

	buf[n] = '\0';
	conn = atoi(buf);

	snprint(buf, sizeof(buf), "url %S", u->src.r);
	if(fswrite(cfid, buf, strlen(buf)) < 0){
		fprint(2, "write: %s: %r\n", buf);
    Err:
		fsclose(cfid);
		return -1;
	}
	if(u->method==HPost && u->post.r != nil){
		snprint(buf, sizeof(buf), "%d/postbody", conn);
		fid = fsopen(webfs, buf, OWRITE);
		if(fid == nil){
			fprint(2, "urlopen: bad query: %s: %r\n", buf);
			goto Err;
		}
		snprint(buf, sizeof(buf), "%S", u->post.r);
		if(fswrite(fid, buf, strlen(buf)) < 0)
			fprint(2, "urlopen: bad query: %s: %r\n", buf);

		fsclose(fid);
	}
	snprint(buf, sizeof(buf), "%d/body", conn);
	fd = fsopenfd(webfs, buf, OREAD);
	if(fd < 0){
		fprint(2, "open: %S: %r\n", u->src.r);
		goto Err;
	}
	u->ctype = getattr(conn, "contenttype");
	u->act = getattr(conn, "parsed/url");
	if(u->act.nr == 0)
		copyrunestr(&u->act, &u->src);

	fsclose(cfid);
	return fd;
}

/* this is a HACK */
Rune *
urlcombine(Rune *b, Rune *u)
{
	Rune *p, *q, *sep, *s;
	int restore;

	if(u == nil)
		error("urlcombine: u == nil");

	if(validurl(u))
		return erunestrdup(u);

	if(b==nil || !validurl(b))
		error("urlcombine: b==nil || !validurl(b)");

	if(runestrncmp(u, (Rune[]){'/','/',0}, 2) == 0){
		q =  runestrchr(b, L':');
		return runesmprint("%.*S:%S", (int)(q-b), b, u);
	}
	p = runestrstr(b, (Rune[]){':','/','/',0});
	sep = (Rune[]){0};
	if(*u ==L'/')
		q = runestrchr(p+3, L'/');
	else if(*u==L'#' || *u==L'?')
		q = runestrrchr(p+3, L'/');
	else{
		sep = (Rune[]){'/',0} ;
		restore = 0;
		s = runestrchr(p+3, L'?');
		if(s != nil){
			*s = '\0';
			restore = 1;
		}
		q = runestrrchr(p+3, L'/');
		if(restore)
			*s = L'?';
	}
	if(q == nil)
		return runesmprint("%S%S%S", b, sep, u);
	else
		return runesmprint("%.*S%S%S", (int)(q-b), b, sep, u);
}
